#ifndef INCLUDED_IntersectionByTriangles
#define INCLUDED_IntersectionByTriangles

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#include "stdmacros.h"
#include "Array.h"
#include "Array2D.h"
#include "LinkedList.h"
#include "ManifoldPointStruct.3Space.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
typedef Array2D<rPoint3> Array2D_rPoint3;
typedef Array2D<boole> Array2D_boole;
typedef LinkedList<AlphaCoord> LinkedList_AlphaCoord;

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

struct ClusterLatticeCluster;
typedef LinkedList<ClusterLatticeCluster> LinkedList_ClusterLatticeCluster;

///////////////////////////////////////////////////////////////////////
struct MCMC_State_Detailed {
  int				size;
  Array<int>			t,k;
  Array<real>			tt,kk;		// meniscus
  Array<int>			itt,ikk;	// sub lattice coords
  
  MCMC_State_Detailed(const int _size=0)
    { size=_size;
      t.SetSize(size); k.SetSize(size);
      tt.SetSize(size); kk.SetSize(size);
      itt.SetSize(size); ikk.SetSize(size);
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// structure so a group of tightly-related functions can share common variables easily
//////////////////////////////////////////////////////////////////////
struct IntersectionByTriangles {

  //::: info that must be passed in to call the algorithm
  const Array<ViewPairInfo>&	pairinfo;		//holds epipolar geometry info for each pair in use
  const Array<Transform>&	T0i_proj;		//this combined with E form all camera matrices in common projective basis
  const Array<rPoint3>&		E;
  const AlphaCoord&		__true_alpha;
  CONST AlphaCoord		true_alpha;
  const Transform&		normalizer;
  Array2D<rPoint3>*		GLOB_Kpts;		//simple kludge used in outputting affine reconstructions
  Transform* const		pInternal;		//allows true internal params to be passed in for testing
  rPoint2			image_size;		//image size is used for showing camera fulcrums and for sanity checking calibration

  //::: local constants
  const int			CON_n_manifolds;	//just pairinfo.size, renamed to make code clearer

  //::: local variables
//  Array<boole>			bManifoldOn;
  
  //::: local variables for manifold coordinate grids
  real				theta_min, theta_max;	// range for theta
  real				kappa_min, kappa_max;	// range for kappa
  int				n_theta;		// steps for theta
  int				n_kappa;		// steps for kappa
  Array2D<real>			Theta;			// theta values for grid lattice for each manifold
  Array2D<real>			Kappa;			// kappa values for grid lattice for each manifold
  Array<Array2D_rPoint3>	lattice;		// lattice points of the grid imposed on each manifold
  Array<Array2D_boole>		b_lattice;		// whether a given lattice point has meaning (or is out of bounds)

  //::: enumerations
  enum { METHOD_MCMC, METHOD_TRI, METHOD_MODC, METHOD_KRUP };

  //::: methods / local functions
  void Run(const String&, const Array2D<boole>&);
  void RansacIteration(rPoint3&, Transform&, int&, real&, const int, const Array<int>&, const Array<LinkedList_ClusterLatticeCluster>&);
  void OutputResults(const rPoint3& best_ans, const Transform& best_K);
  void PerformIntersectionByTriangles(LinkedList<rPoint3>&, const Array<int>&, const Array<LinkedList_ClusterLatticeCluster>&, const boole);

  real Likelihood(const MCMC_State_Detailed&);
  void PerformIntersectionByMcmc(LinkedList<rPoint3>&, const Array<int>&, const boole);

  void PerformIntersectionByModc(LinkedList<rPoint3>&, const Array<int>&, const boole);
  void VisualizeModcTerrain(const rPoint3&, const real&, const real&, const int, const Array<rPoint3>&, const rPoint3&, const Array<int>&);
  void GradientDescent(rPoint3&, const Array<int>&);
  void BruteForceSearch(rPoint3&, const real&, const int, const Array<int>&);

  IntersectionByTriangles(const Array<ViewPairInfo>&, const Array<Transform>&,
    const Array<rPoint3>&, const rPoint3&, const Transform&, const rPoint2&,
    Array2D<rPoint3>*, Transform* const _pInternal=NULL);
  ~IntersectionByTriangles();
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

struct IntersectionByTrianglesError {
  String	msg;
  int		error_code;

  IntersectionByTrianglesError(const String& _msg, const int code=-1) { msg=_msg; error_code=code; }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// triangle intersection routines; used by MCMC intersection

// placed here so outside functions can call tri_tri_intersect routine
//////////////////////////////////////////////////////////////////////
struct ClusterTriangle {
  rPoint3			a,b,c;			// vertices in search space coords
  
  rPoint3			u, v, n, u_cr_v;	// precomputed values for faster triangle intersection testing
//  real				a_th, a_ka;	// theta and kappa coords for vertex a
//  real				b_th, b_ka;	// used to estimate manifold coords of intersection pt
//  real				c_th, c_ka;	//

  void Init(const rPoint3& A, const rPoint3& B, const rPoint3& C);

  ClusterTriangle() { }
  ClusterTriangle(const rPoint3& A, const rPoint3& B, const rPoint3& C) { Init(A,B,C); }
};

//////////////////////////////////////////////////////////////////////
struct ClusterLatticeCluster {
  int				th0, th1;
  int				ka0, ka1;
  rPoint3			lo, hi;		// the bounding box
  boole				b_legal;	// does cluster have at least one legal lattice point?
  LinkedList<ClusterTriangle>	tri;		// triangles for this cluster
  boole				b_tri_exist;	// have tris been generated yet?
};

//////////////////////////////////////////////////////////////////////
struct ClusterPair {
  ClusterLatticeCluster*	a;
  ClusterLatticeCluster*	b;
};

//////////////////////////////////////////////////////////////////////
struct ClusterTrianglePair {
  ClusterTriangle		a, b;
  rPoint3			lo, hi;		// bounding box for pair
};

//////////////////////////////////////////////////////////////////////
struct ClusterTriangleTriplet {
  ClusterTriangle		a, b, c;
};

//////////////////////////////////////////////////////////////////////
typedef LinkedList<ClusterTriangle> LinkedList_ClusterTriangle;

//////////////////////////////////////////////////////////////////////
int tri_tri_intersect(
  const ClusterTriangle&	S,
  const ClusterTriangle&	T
);

//////////////////////////////////////////////////////////////////////
void IntersectionOf3Triangles(
  boole&			b_intersect,		// do the tris intersect?
  rPoint3&			P,			// intersection point
  const ClusterTriangle&	A,
  const ClusterTriangle&	B,
  const ClusterTriangle&	C
);

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

#endif
