#include "MultiCameraWorkbench.h"
#include "helpers.h"
#include "MiscUtility.h"
#include "FundamentalMatrix.h"
#include "VrmlFile.h"
#include "RM_SameCameraManip.h"
#include "KltFeatureTable.h"
#include "TestRunDataStruct.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
rPoint3 random_point()
{
  return rPoint3(randpm()*500,randpm()*500,randpm()*500);
}

///////////////////////////////////////////////////////////////////////
rPoint3 random_direction()
{
  real a1=rand01()*PI2;
  real a2=(rand01()-0.5)*PI;
  real c=abs(cos(a2));
  return rPoint3(c*cos(a1),c*sin(a1),sin(a2));
}

///////////////////////////////////////////////////////////////////////
Transform random_transform()
{
  Transform T;
  FORi(3) FORj(3) T[i][j]=randpm();
  return T;
}

// returned pts are on image plane
///////////////////////////////////////////////////////////////////////
rPoint3 ProjectIntoCamera(const rMatrix& CM, const rPoint3& P)
{
  static rMatrix p(4,1), q;
  p[0][0]=P.x;    p[1][0]=P.y;    p[2][0]=P.z;    p[3][0]=1;
  q=CM*p;
  if (q[2][0]==0) EXIT("unexpected point at infinity");
  real recip=1.0/abs(q[2][0]);
  return rPoint3(q[0][0]*recip, q[1][0]*recip, q[2][0]*recip);
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
void ExtractHinf(
  Transform&			Hinf,
  const rMatrix&		A,
  const rMatrix&		B
)
{
  Transform TWA(A[0][0],A[0][1],A[0][2],A[1][0],A[1][1],A[1][2],A[2][0],A[2][1],A[2][2]);
  Transform TWB(B[0][0],B[0][1],B[0][2],B[1][0],B[1][1],B[1][2],B[2][0],B[2][1],B[2][2]);
  Hinf=TWB*(TWA.Inverse());
}

/*\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
-----------------------------------------------------------------------
find 3D vector A such that Hinf == M + eL * A'
-----------------------------------------------------------------------
\*/////////////////////////////////////////////////////////////////////
void FindVectorA(
  rVector&			VectorA,
  const Transform&		Hinf,
  const Transform&		M,
  const rPoint3&		eL
)
{
  rMatrix Q(9,4);
  Q.FillWith(0);
  Q[0][0]=M[0][0];  Q[1][0]=M[0][1];  Q[2][0]=M[0][2];
  Q[3][0]=M[1][0];  Q[4][0]=M[1][1];  Q[5][0]=M[1][2];
  Q[6][0]=M[2][0];  Q[7][0]=M[2][1];  Q[8][0]=M[2][2];

  Q[0][1]=eL.x;  Q[3][1]=eL.y;  Q[6][1]=eL.z;
  Q[1][2]=eL.x;  Q[4][2]=eL.y;  Q[7][2]=eL.z;
  Q[2][3]=eL.x;  Q[5][3]=eL.y;  Q[8][3]=eL.z;
  
  rMatrix H(9,1);
  H[0][0]=Hinf[0][0];  H[1][0]=Hinf[0][1];  H[2][0]=Hinf[0][2];
  H[3][0]=Hinf[1][0];  H[4][0]=Hinf[1][1];  H[5][0]=Hinf[1][2];
  H[6][0]=Hinf[2][0];  H[7][0]=Hinf[2][1];  H[8][0]=Hinf[2][2];
  
  rMatrix tranQ=Q.TransposeOf();
  VectorA=(tranQ*Q).InverseOf()*tranQ*H;
  VectorA[1]/=(VectorA[0]);
  VectorA[2]/=(VectorA[0]);
  VectorA[3]/=(VectorA[0]);
  VectorA[0]=VectorA[1];
  VectorA[1]=VectorA[2];
  VectorA[2]=VectorA[3];
  VectorA[3]=1.0;
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
void CreateScene(Array<rPoint3>& W, const int size)
{
  W.SetSize(size);
  FOREACHi(W) W[i]=random_point();
}

///////////////////////////////////////////////////////////////////////
void CreateScene_Cube_2(Array<rPoint3>& W)
{
//  const int width=10;
  const int width=6;
  W.SetSize(width*width*3);
  int k=0;
  FORi(width) FORj(width) W[k++]=rPoint3(i,j,0);
  FORi(width) FORj(width) W[k++]=rPoint3(i,0,j);
  FORi(width) FORj(width) W[k++]=rPoint3(0,i,j);
  FORi(W.size) W[i]*=1.0/width;
}

// this isn't a cube -- it's 2 faces of a cube like in Pollefeys' thesis;
// allows for cameras to be more spread out
// features are in range [0,1]
///////////////////////////////////////////////////////////////////////
void CreateScene_Cube_3(Array<rPoint3>& W)
{
  const int width=7;
  W.SetSize(width*width*2);
  int k=0;
  FORiIN(1,width) FORjIN(1,width) W[k++]=rPoint3(i,0,j);
  FORiIN(1,width) FORjIN(1,width) W[k++]=rPoint3(0,i,j);
  FORi(W.size) W[i]*=1.0/width;
}

///////////////////////////////////////////////////////////////////////
void CreateScene_Cube(Array<rPoint3>& W)
{
  W.SetSize(300);
  int k=0;
  FORi(10) FORj(10) W[k++]=rPoint3(i,j,0);
  FORi(10) FORj(10) W[k++]=rPoint3(i,0,j);
  FORi(10) FORj(10) W[k++]=rPoint3(0,i,j);
  FORi(W.size) W[i]+=rPoint3(3,3,1);
}

///////////////////////////////////////////////////////////////////////
void AddACamera(
  Array<rPoint3>&		A,
  rMatrix&			CMA,
  const Array<rPoint3>&		W,
  const rPoint3&		centerW
)
{
  A.SetSize(W.size);
  CMA.SetSize(3,4);

  Transform TWA=random_transform();
//  TWA=Transform_Id;

  FORi(3) FORj(3) CMA[i][j]=TWA[i][j];
  rPoint3 a=TWA*centerW*(-1);
  CMA[0][3]=a.x;  CMA[1][3]=a.y;  CMA[2][3]=a.z;

  FOREACHi(W) A[i]=ProjectIntoCamera(CMA,W[i]);
}

///////////////////////////////////////////////////////////////////////
Transform MakeRandomRealInternalParams_2(const rPoint2& CON_simulated_image_size)
{
  Transform TWK;

//rPoint3 lo(9999,9999,9999), hi(-1,-1,-1);
//INFINITELOOP{

  rPoint2 princ_pt(0.5+randpm()*0.15, 0.5+randpm()*0.15);	//range: up to 15% of half-diagonal length
  TWK.FillWith(0);
  TWK[0][0]=1+randpm()*0.2;
  TWK[1][1]=TWK[0][0]+randpm()*1.0/20;
real lens_width_factor=0.2;                   // "wide lens"
//real lens_width_factor=0.6;                     // "medium lens"
TWK[0][0]*=lens_width_factor;
TWK[1][1]*=lens_width_factor;
  TWK[0][1]=randpm()*1.0/25;
//TWK[0][1]=0;
  TWK[0][2]=princ_pt.x;
  TWK[1][2]=princ_pt.y;
  TWK[2][2]=1;
  TWK=Transform(CON_simulated_image_size.x,CON_simulated_image_size.y,1)*TWK;

/*
real xx=TWK[0][0]/TWK[1][1];
real yy=TWK[0][2]/TWK[1][2]*CON_simulated_image_size.y/CON_simulated_image_size.x;
lo.x=min(xx,lo.x);
hi.x=max(xx,hi.x);
lo.y=min(yy,lo.y);
hi.y=max(yy,hi.y);
//printf("%15e   0.85  1.15\n",TWK[0][0]/TWK[1][1]);
//printf("%15e   0.90  1.10\n\n",TWK[0][2]/TWK[1][2]*CON_simulated_image_size.y/CON_simulated_image_size.x);
lo.Dump("lo");
hi.Dump("hi");
}
*/

  return TWK;
}

///////////////////////////////////////////////////////////////////////
Transform MakeRandomRealInternalParams()
{
  Transform TWK;
  real focal_length=rand01()*.4+0.7;			//range: .4 - 1.1
  real skew_angle=randpm()*5*PI/180;			//range: +/- 5 degrees
  real y_length=1+randpm()*0.05;			//range: 1+/-0.05
  rPoint2 princ_pt(randpm()*0.5*0.15, randpm()*0.5*0.15);	//range: up to 15% of half-diagonal length
  TWK.FillWith(0);
  TWK[0][0]=1;
  TWK[0][1]=sin(skew_angle)*y_length;
  TWK[1][1]=cos(skew_angle)*y_length;
  TWK[0][2]=princ_pt.x;
  TWK[1][2]=princ_pt.y;
  TWK[2][2]=focal_length;
  return TWK;
}

///////////////////////////////////////////////////////////////////////
void AssembleCameraMatrix(
  rMatrix&			CM,
  const Transform&		internal,
  const Transform&		external,
  const rPoint3&		center_W
)
{
  Transform TWK=internal*external;

  CM.SetSize(3,4);
  FORi(3) FORj(3) CM[i][j]=TWK[i][j];

  rPoint3 oc=TWK*center_W*(-1);
  CM[0][3]=oc.x;  CM[1][3]=oc.y;  CM[2][3]=oc.z;
}

///////////////////////////////////////////////////////////////////////
void MakeRandomRealCamera(rMatrix& CM)
{
  Transform internal=MakeRandomRealInternalParams();
  rPoint3 axis=random_direction();
  real rot=rand01()*PI2;
  AssembleCameraMatrix(CM,internal,Transform::Rotation(axis,rot),random_point());
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
real ProjectiveReconBadness(
  const rPoint3&		alpha,
  const Transform&		M,
  const rPoint3&		eL,
  const Array<rPoint3>&		_pts1,
  const Array<rPoint3>&		_pts2,
  const Transform&		F,
  const int			phase,
  String*			pName=NULL
)
{
  static boole bWarned=FALSE;
  Transform MM=M+Transform::Outer(eL,alpha);

  const int CON_skip=5;
  if ((!bWarned)&&(CON_skip>1)) {
    SCREENPUT ALERT("not using all pts in evaluating initial reconstruction");
    bWarned=TRUE;
  }

  Array<rPoint3> pts1, pts2;
  if (phase==1) {
    pts1.MakeCopyOf(_pts1);
    pts2.MakeCopyOf(_pts2);
  } else {
    pts1.SetSize(_pts1.size/CON_skip);
    pts2.SetSize(_pts2.size/CON_skip);
    FORi(pts1.size) pts1[i]=_pts1[CON_skip*i];
    FORi(pts2.size) pts2[i]=_pts2[CON_skip*i];
  }

  Array<rPoint3> RR;
  MiscUtility::AffineTriangulate(RR,MM.NormalizationOf(),F,pts1,pts2);

  int nPositive=0;
  FORi(RR.size) if (RR[i].z>0) ++nPositive;
//printf("%f\n",real(nPositive)/RR.size);
//S.Dump("S");
  if ((phase==2)&&(nPositive<RR.size)) return 1e99;

  rMatrix V,S;
  rPoint3 median;
  MiscUtility::FindPrincipalComponentsOf3DPointCloud(V,S,median,RR);
  
if (pName!=NULL) {
  VrmlFile::Open(*pName);
  VrmlFile::OutputPoints(RR,RGB3Byte_green);
  Array<rPoint3> lines(6);
  lines[0]=median;
  lines[2]=median;
  lines[4]=median;
  lines[1]=rPoint3(V[0][0],V[0][1],V[0][2])*50;
  lines[3]=rPoint3(V[1][0],V[1][1],V[1][2])*50;
  lines[5]=rPoint3(V[2][0],V[2][1],V[2][2])*50;
  VrmlFile::OutputLines(lines,RGB3Byte_red);
  VrmlFile::Close();
}

  real x=abs(S[0][0]);
//  real y=abs(S[1][1]);
  real z=abs(S[2][2]);
  if (phase==1) return -SQ(z/x)-100*real(nPositive)/RR.size;

  return -SQ(z/x);	//assumes x is max, z is min singular value
}

#ifdef NOTDEFINED
///////////////////////////////////////////////////////////////////////
rPoint3 ProjectiveReconBadness2(
  const rPoint3&		alpha,
  const Transform&		M,
  const rPoint3&		eL,
  const Array<rPoint3>&		pts1,
  const Array<rPoint3>&		pts2,
  const Transform&		F
)
{
  Transform MM=M+Transform::Outer(eL,alpha);

  Array<rPoint3> RR;
  MiscUtility::AffineTriangulate(RR,MM.NormalizationOf(),F,pts1,pts2);

  rMatrix V,S;
  rPoint3 median;
  MiscUtility::FindPrincipalComponentsOf3DPointCloud(V,S,median,RR);
  
  int nPositive=0;
  FORi(RR.size) if (RR[i].z>0) ++nPositive;

  real x=abs(S[0][0]);
  real y=abs(S[1][1]);
  real z=abs(S[2][2]);

  rPoint3 ans(SQ(z/x),SQ(y/x),SQ(z/y));
  if (nPositive<RR.size) return ans+rPoint3(1e99,1e99,1e99);
  return ans;
}

// used with Levenberg-Marquardt; didn't work because error function to bumpy?
///////////////////////////////////////////////////////////////////////
MultiCameraWorkbench* PRBerror_globals=NULL;
real PRBerror(const rVector& x, const int i)
{
  const int one=0, two=1;
  rPoint3 A(x[0],x[1],x[2]);
  rPoint3 err=rPoint3(-1,-1,-1)+ProjectiveReconBadness2(A,PRBerror_globals->M[one][two],
    PRBerror_globals->eL[one][two],PRBerror_globals->Kpts[one],
    PRBerror_globals->Kpts[two],PRBerror_globals->F[one][two]);
  if (i==-1) return SQ(err.Norm());
  else switch (i) {
    case 0: return err.x;
    case 1: return err.y;
    case 2: return err.z;
    default:  EXIT("bad bad bad");  return 0;
  }
}
#endif


#include "Image3ByteChannels.h"

///////////////////////////////////////////////////////////////////////
void MultiCameraWorkbench::MakeInitialReconstructionRounder(
  const int			one,
  const int			two
)
{
  //::: find alpha that puts all world pts in front of cameras one and two
  int loop=1;
  rPoint3 alpha, bestalpha;
  real besterr=1e99;
  INFINITELOOP {
    alpha=rPoint3(randpm()*10,randpm()*10,randpm()*10);		// magic numbers
    real error=ProjectiveReconBadness(alpha,M[one][two],eL[one][two],Kpts[one],Kpts[two],F[one][two],1);
    if (error<besterr) {
      besterr=error;
      bestalpha=alpha;
//      alpha.Dump("alpha");
      SCREENPUT printf("err=%20.16f\n",error);
    }
    if (besterr<=-100) break;

    ++loop;
  }

  alpha=bestalpha;

  //::: change M[one][two] until R[one][two] has nice, filled-out shape
  //::: exhaustive search
  FILE* file=fopen("precomputed.bestalpha","rb");
  if (file) {
    bestalpha.ReadFromFile(file);
    fclose(file);
  } else {
    VolumeStruct region;
    region.min=rPoint3(-5,-5,-5);
    region.max=rPoint3(5,5,5);
    real step=1;	//2 might work
    real shrink=0.5;
    besterr=1e99;	//magic number
    rPoint3 A;
    real limit=100;
///  LinkedList<rPoint3> goodAs;	// for cheirality plot
    FORi(14) {
    SCREENPUT printf("------------- loop = %d --------------\n",i);
    for (A.x=region.min.x; A.x<=region.max.x; A.x+=step) {
      SCREENPUT printf("%f\n",A.x);
      for (A.y=region.min.y; A.y<=region.max.y; A.y+=step)
        for (A.z=region.min.z; A.z<=region.max.z; A.z+=step) {
Transform MM=M[one][two]+Transform::Outer(eL[one][two],A);
real dakfds;
if ((dakfds=abs(abs(MM.NormalizationOf().Determinant())-0.19245))>limit) continue;
          real error=ProjectiveReconBadness(A,M[one][two],eL[one][two],Kpts[one],Kpts[two],F[one][two],2);
///      if (error<100) goodAs.Add(new rPoint3(A));
          if (error<besterr) {
            besterr=error;
            bestalpha=A;
  	    limit=dakfds*1.1;
//            A.Dump("new best alpha");
            SCREENPUT printf("err=%20.16f    limit=%e\n",error,limit);
            if (besterr<-0.3) goto SEARCH_DONE;
//            if (besterr<-0.05) goto SEARCH_DONE;
          }
        }
      }
      rPoint3 span=region.max-region.min;
      region.min=bestalpha-span*(shrink/2);
      region.max=bestalpha+span*(shrink/2);
      step*=shrink;
    }

SEARCH_DONE:
//    FILE* file=fopen("precomputed.bestalpha","wb");
    FILE* file=fopen("XXXprecomputed.bestalpha","wb");
    bestalpha.WriteToFile(file);
    fclose(file);

///Array<rPoint3> _goodAs;
///_goodAs.MakeCopyOf(goodAs);
///VrmlFile::OutputPoints("cheiralplot.wrl",_goodAs,RGB3Byte_green);

  }

  SCREENPUT bestalpha.Dump("bestalpha");
  M[one][two]=M[one][two]+Transform::Outer(eL[one][two],bestalpha);
SCREENPUT printf("det = %e\n",(M[one][two].NormalizationOf().Determinant()));
}



// solve qq ~~ cm*QQ for cm
//////////////////////////////////////////////////////////////////////
void FindCameraMatrix_linear(
  rMatrix&			cm,
  const Array<rPoint3>&		QQ,		// world coords
  const Array<rPoint3>&		qq		// image plane coords
)
{
  ASSERT(QQ.size==qq.size,"err9823598");
  rMatrix A(2*QQ.size,12);
  real Qw=4;  real qz=250;	//kludgy attempt at stabilizing the linear algorithm; these are homogeneous depth values
  				//final matrix must be scaled to eliminate these factors
  int i,k;
  for (i=0, k=0; i<QQ.size; ++i) {
    const rPoint3& Q=QQ[i];
    const rPoint2& q=qq[i];
    A[k][0]= Q.x*q.y;
    A[k][1]= Q.y*q.y;
    A[k][2]= Q.z*q.y;
    A[k][3]=  Qw*q.y;
    
    A[k][4]=-Q.x*q.x;
    A[k][5]=-Q.y*q.x;
    A[k][6]=-Q.z*q.x;
    A[k][7]= -Qw*q.x;
    
    A[k][8]=A[k][9]=A[k][10]=A[k][11]=0;
    ++k;

    A[k][0]= Q.x*qz;
    A[k][1]= Q.y*qz;
    A[k][2]= Q.z*qz;
    A[k][3]=  Qw*qz;
    
    A[k][4]=A[k][5]=A[k][6]=A[k][7]=0;

    A[k][8] =-Q.x*q.x;
    A[k][9] =-Q.y*q.x;
    A[k][10]=-Q.z*q.x;
    A[k][11]= -Qw*q.x;
    ++k;
  }

  rMatrix B=A.TransposeOf()*A;
  A.FreeMem();
  rMatrix U,S,V;
  B.full_Svd(U,S,V);

  cm.SetSize(3,4);
  int r=11;
  cm[0][0]=V[r][0];  cm[0][1]=V[r][1];  cm[0][2]=V[r][2];  cm[0][3]=V[r][3];
  cm[1][0]=V[r][4];  cm[1][1]=V[r][5];  cm[1][2]=V[r][6];  cm[1][3]=V[r][7];
  cm[2][0]=V[r][8];  cm[2][1]=V[r][9];  cm[2][2]=V[r][10]; cm[2][3]=V[r][11];

//undo the stabilization scaling
rMatrix Wsc(4,4); Wsc.MakeIdentity(); Wsc[3][3]=Qw;
rMatrix Zsc(3,3); Zsc.MakeIdentity(); Zsc[2][2]=1.0/qz;
cm=Zsc*cm*Wsc;

/*  
  Transform Hinf(cm[0][0],cm[0][1],cm[0][2],cm[1][0],cm[1][1],cm[1][2],cm[2][0],cm[2][1],cm[2][2]);
  rPoint3 u(cm[0][3],cm[1][3],cm[2][3]);
  ((Hinf.Inverse()*u)*(-1)).Dump("optical center");
  cm.Dump("camera matrix");
*/
}

///////////////////////////////////////////////////////////////////////
void MultiCameraWorkbench::CalcAllVarsFromCameraMatrices(
  const int			one,
  const int			two
)
{
  //::: find f-mats
  F.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) F[i][j]=FundamentalMatrix::Find_known_cameras(CM[i],CM[j]);

  //::: find epipoles
  eR.SetSize(CON_ncameras,CON_ncameras);
  eL.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) FundamentalMatrix::FindEpis(eR[i][j],eL[i][j],F[i][j]);

  //::: find any M such that F == [eB]_x*M
  M.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) M[i][j]=Transform::Cross(eL[i][j])*F[i][j];  //ERROR!!!  tell Ziss!?
    if (i<j) FundamentalMatrix::FindNonhomogeneous(M[i][j],F[i][j],eL[i][j]);

//  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) { F[i][j].DumpNormalized(); (Transform::Cross(eL[i][j])*M[i][j]).DumpNormalized(); }

  //::: find true Hinf's
  Hinf.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) ExtractHinf(Hinf[i][j],CM[i],CM[j]);

  //::: find alpha such that Hinf == M + e_B*alpha'
  VectorA.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) FindVectorA(VectorA[i][j],Hinf[i][j],M[i][j],eL[i][j]);
    // NOTE: must apply normalizing transform to camera matrices for FindVectorA to work

  //::: store vectorA in convenient form
  alpha.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j)
    alpha[i][j]=AlphaCoord(VectorA[i][j][0],VectorA[i][j][1],VectorA[i][j][2]);

  //::: pairwise projective reconstructions
  R.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) MiscUtility::AffineTriangulate(R[i][j],M[i][j],F[i][j],Kpts[i],Kpts[j]);

  SCREENPUT VrmlFile::OutputPoints("output.initial.projective.recon.wrl",R[one][two],RGB3Byte_orange);

  //::: collineations between projective reconstructions
  T.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if ((i<j)&&((i!=0)||(j!=1))) MiscUtility::Calc4DCollineation(T[i][j],R[i][j],R[0][1]);
    if ((i<j)&&((i!=one)||(j!=two))) MiscUtility::Calc4DCollineation(T[i][j],R[one][two],R[i][j]);

  //::: find true parameterizations; store "view pairs" separately
  k.SetSize(CON_pairs);
  th.SetSize(CON_pairs);
  g.SetSize(CON_pairs);
  int ind=0;
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j) {
    FindParameterizationOfHMatrix(k[ind],th[ind],g[ind],internal,
      Hinf[i][j],F[i][j],M[i][j],eL[i][j],TRUE);
    ++ind;
  }
}

///////////////////////////////////////////////////////////////////////
void MultiCameraWorkbench::MakeInitialReconstructionRounder_2(
  const int			one,
  const int			two
)
{
  //::: find alpha that puts all world pts in front of cameras one and two
  int loop=1;
  rPoint3 alpha, bestalpha;
  real besterr=1e99;
  INFINITELOOP {
    alpha=rPoint3(randpm()*10,randpm()*10,randpm()*10);		// magic numbers
    real error=ProjectiveReconBadness(alpha,M[one][two],eL[one][two],Kpts[one],Kpts[two],F[one][two],1);
    if (error<besterr) {
      besterr=error;
      bestalpha=alpha;
//      alpha.Dump("alpha");
      printf("err=%20.16f\n",error);
    }
    if (besterr<=-100) break;

    ++loop;
  }

  alpha=bestalpha;


//  Transform MM=M[one][two];
  Transform MM=M[one][two]+Transform::Outer(eL[one][two],alpha);
  Array<rPoint3> P;
  MiscUtility::AffineTriangulate(P,MM.NormalizationOf(),F[one][two],Kpts[one],Kpts[two]);

  //::: find median, covariance matrix, and principal components for projective reconstruction
  rMatrix Q(3,P.size);
  rPoint3 median=rPoint3(0,0,0);
  FORi(P.size) median+=P[i];
  median/=P.size;
  FORi(P.size) { Q[0][i]=P[i].x-median.x; Q[1][i]=P[i].y-median.y; Q[2][i]=P[i].z-median.z; }
  rMatrix QQ=Q*(Q.TransposeOf());	// covariance matrix
  rMatrix U,S,V;
  QQ.full_Svd(U,S,V);
  V=V.TransposeOf();

  //::: transform samples so major axii are principal components
  Transform T(&V[0]);
  T.Invert();
  FORi(P.size) P[i]=T*(P[i]-median);

  //::: find standard deviation in each axis direction
  rPoint3 med2(0,0,0);
  FORi(P.size) med2+=P[i];
  med2/=P.size;
  rPoint3 sd(0,0,0);
  FORi(P.size) {
    sd.x+=SQ(P[i].x-med2.x);
    sd.y+=SQ(P[i].y-med2.y);
    sd.z+=SQ(P[i].z-med2.z);
  }
  sd.x=sqrt(sd.x/P.size);
  sd.y=sqrt(sd.y/P.size);
  sd.z=sqrt(sd.z/P.size);
SCREENPUT sd.Dump("sd");
/*
  //::: scale width of point set in each axis direction so that width is 1
  rPoint3 max, min;
  max=min=M[0];
  FORi(M.size) {
    const rPoint3& p=M[i];
    if (p.x>max.x) max.x=p.x;
    if (p.y>max.y) max.y=p.y;
    if (p.z>max.z) max.z=p.z;
    if (p.x<min.x) min.x=p.x;
    if (p.y<min.y) min.y=p.y;
    if (p.z<min.z) min.z=p.z;
  }
*/
  //::: scale each axis so that standard deviation (or max width) is 1 in each direction
  Transform scale=Transform(1.0/sd.x,1.0/sd.y,1.0/sd.z);
//  Transform scale=Transform(1.0/(max.x-min.x),1.0/(max.y-min.y),1.0/(max.z-min.z));
  FORi(P.size) P[i]=scale*P[i];
//VrmlFile::OutputPoints("dsalfkds.wrl",P,RGB3Byte_green);
//EXIT();

  //::: fit each camera matrix to rounder construction
  CM.SetSize(CON_ncameras);
  FORi(CM.size) FindCameraMatrix_linear(CM[i],P,Kpts[i]);

  //::: create final construction using all views at once
  MiscUtility::Triangulate(P,CM,Kpts);

  //::: fit each camera matrix to final construction
  FORi(CM.size) FindCameraMatrix_linear(CM[i],P,Kpts[i]);
  
  //::: transform CM's and recon so that camera one is [I 0]
  {
  rMatrix Y(4,4);
  Transform YY;
  FORi(3) FORj(3) YY[i][j]=CM[one][i][j];
  YY=YY.Inverse();
  FORi(3) FORj(3) Y[i][j]=YY[i][j];
  rPoint3 y(CM[one][0][3],CM[one][1][3],CM[one][2][3]);
  y=(YY*y)*(-1);
  Y[0][3]=y.x;
  Y[1][3]=y.y;
  Y[2][3]=y.z;
  Y[3][3]=1;
  FORi(CM.size) CM[i]=CM[i]*Y;
  rMatrix X;
  X=Y.InverseOf();
  rVector Q(4);
  Q[3]=1;
  FORi(P.size) {
    Q[0]=P[i].x;
    Q[1]=P[i].y;
    Q[2]=P[i].z;
    Q=X*Q;
    Q/=Q[3];
    P[i]=rPoint3(Q[0],Q[1],Q[2]);
  }
  CM.Dump();
  }
  
  //::: recalc fmats and all related terms again (as if fmats loaded from a file)
  CalcAllVarsFromCameraMatrices(one,two);
}


///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

//#define NEWPATH

// constructor for synthetic data
///////////////////////////////////////////////////////////////////////
MultiCameraWorkbench::MultiCameraWorkbench(
  const int			_CON_ncameras,
  const int			_CON_nworldpts,
  const real&			_CON_noiseradius,
  const int			one,
  const int			two,
  const real&			CON_min_retinal_obj_size,
  const iPoint2&		_image_size
) :
CON_ncameras(_CON_ncameras),
CON_nworldpts(_CON_nworldpts),
CON_noiseradius(_CON_noiseradius),
CON_pairs(CON_ncameras*(CON_ncameras-1)/2),
CON_simulated_image_size(_image_size)
{
//  CreateScene(W, CON_nworldpts);
  CreateScene_Cube_3(W);

  Kpts.SetSize(CON_ncameras,W.size);  
  CM.SetSize(CON_ncameras);
  Twi.SetSize(CON_ncameras);

/*
  //::: use this block for random cameras
  FOREACHi(CM) AddACamera(Kpts[i],CM[i],W,random_point());
*/

/*
  //::: use this block for identical cameras in general config
  internal=MakeRandomRealInternalParams();
  FOREACHi(CM) {
    rPoint3 axis=random_direction();
    real rot=randpm()/5*PI2;
    Twi[i]=internal*Transform::Rotation(axis,rot);
    if (i==0) AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),rPoint3(0,0,0));
    else      AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),random_direction()*0.1);
    FOREACHj(W) Kpts[i][j]=ProjectIntoCamera(CM[i],W[j]);

    if (CON_noiseradius==0)
      FOREACHj(W) rPoint3(randpm(),randpm(),0);			//for using up random numbers
    else
      FOREACHj(W) Kpts[i][j]+=rPoint3(randpm()*CON_noiseradius,randpm()*CON_noiseradius,0);	//for adding noise
  }
*/

  //::: use this block for identical cameras all facing complete world point set
  Array2D<rPoint3> noisefree_Kpts;
  noisefree_Kpts.SetSize(CON_ncameras,W.size);
  internal=MakeRandomRealInternalParams_2(rPoint2(CON_simulated_image_size.x, CON_simulated_image_size.y));
  SCREENPUT { printf("trying to generate cameras...");  fflush(stdout); }
  FOREACHi(CM) {
    try_again:
      rPoint3 axis=random_direction();
///      real rot=randpm()/5*PI2;
      real rot=randpm()*PI2;
      Twi[i]=internal*Transform::Rotation(axis,rot);
///      if (i==0) AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),rPoint3(0,0,0));
///      else      AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),random_direction()*0.1);
//      AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),rPoint3(rand01(),rand01(),rand01())*2+rPoint3(1.1,1.1,1.1));	// for Cube_2() world
//      AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),rPoint3(rand01(),rand01(),randpm()*2)*2+rPoint3(1.1,1.1,0));	// for Cube_3() world
      AssembleCameraMatrix(CM[i],internal,Transform::Rotation(axis,rot),rPoint3(rand01()*2,rand01()*2,rand01()*1)+rPoint3(0.3,0.3,0));	// for Cube_3() world, wide lens

      Kpts[i].SetSize(W.size);
      noisefree_Kpts[i].SetSize(W.size);
      rPoint2 centroid(0,0);
      Array<rPoint2> ppp(W.size);
      FOREACHj(W) {
        rPoint3& p=Kpts[i][j];
        p=ProjectIntoCamera(CM[i],W[j]);
	if (p.z<=0) goto try_again;
	rPoint3 pp=p/p.z;
	if (!within(pp.x,0,CON_simulated_image_size.x)) goto try_again;
	if (!within(pp.y,0,CON_simulated_image_size.y)) goto try_again;
        centroid.x+=pp.x;
        centroid.y+=pp.y;
        ppp[j].x=pp.x;
        ppp[j].y=pp.y;
      }

      //::: make sure object covers enough of image plane
      if (CON_min_retinal_obj_size>0) {
        centroid/=W.size;
        real variance=0;
        FOREACHj(W) {
          rPoint2 pp=ppp[j];
          pp.x=centroid.x-pp.x;
          pp.y=centroid.y-pp.y;
          variance+=pp.x*pp.x+pp.y*pp.y;
        }
        real stddev=sqrt(variance/W.size);
        if (CON_min_retinal_obj_size>stddev) goto try_again;
      }

    noisefree_Kpts[i]=Kpts[i];
    if (CON_noiseradius==0)
      FOREACHj(W) { rPoint3 dum=rPoint3(randpm(),randpm(),0); }		//for using up random numbers
    else
      FOREACHj(W) Kpts[i][j]+=rPoint3(randpm()*CON_noiseradius,randpm()*CON_noiseradius,0);	//for adding noise
  }
  SCREENPUT { printf("done!\n");  fflush(stdout); }

  SCREENPUT {
    //::: show view from camera 0
    Image3ByteChannels img(CON_simulated_image_size.x,CON_simulated_image_size.y);
    img.FillWithColor(RGB3Byte_black);
    FORi(Kpts[0].size) img.DrawSquare(rPoint2(Kpts[0][i].x,Kpts[0][i].y),3,RGB3Byte_red);
    FORi(noisefree_Kpts[0].size) img.DrawSquare(rPoint2(noisefree_Kpts[0][i].x,noisefree_Kpts[0][i].y),3,RGB3Byte_green);
    FORi(Kpts[0].size) img.DrawLine(rPoint2(Kpts[0][i].x,Kpts[0][i].y),rPoint2(noisefree_Kpts[0][i].x,noisefree_Kpts[0][i].y),RGB3Byte_blue);
    img.WriteToFile("dasjfldsk.jpg");
//    EXIT();
  }
  
  //::: make sure camera 0 at origin
  Transform T_met_0;
  FORi(3) FORj(3) T_met_0[i][j]=CM[0][i][j];
  Transform T_0_met=T_met_0.Inverse();
  rPoint3 cent0(CM[0][0][3],CM[0][1][3],CM[0][2][3]);
  cent0=T_0_met*cent0*(-1);
  rMatrix T_shift(4,4);
  T_shift.MakeIdentity();
  T_shift[0][3]=cent0.x;
  T_shift[1][3]=cent0.y;
  T_shift[2][3]=cent0.z;
  FORi(CM.size) CM[i]=CM[i]*T_shift;
  FORi(W.size) W[i]-=cent0;

  //::: use to show initial setup
  Array<rPoint3> nul;
  SCREENPUT VrmlFile::OutputPointsAndCameras("output.starting.scenario.wrl",W,nul,CM,rPoint2(CON_simulated_image_size.x, CON_simulated_image_size.y));
//  EXIT();

  //::: normalize points on camera image planes
////--bad mistake-->  FORi(CON_ncameras) MiscUtility::FindNormalizingTransform(Tinorm[i],Kpts[i]); //must use same transform for each camera!!!
  MiscUtility::FindNormalizingTransform(normalization,Kpts[one]);
//normalization=Transform_Id;
  SCREENPUT internal.DumpNormalized("internal params");  
  internal=normalization*internal;
  SCREENPUT internal.DumpNormalized("normalized internal params");  
  rMatrix rm_normalization(&normalization[0][0],3,3);
  FORi(CON_ncameras) CM[i]=rm_normalization*CM[i];

  FORi(Kpts.rows) FORj(Kpts[i].size)
    Kpts[i][j]=normalization*Kpts[i][j];	//note the z-coord remains the same (at 1); Tinorm must be applied in this order so that each K matrix modified the same way; all Tinorms must be the same

  //::: turn on all pairs
  bPairOn.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras) bPairOn[i][j]=FALSE;
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j) bPairOn[i][j]=TRUE;



#ifdef NEWPATH
//::: new path
ALERT("NEW PATH");

  //::: find good initial reconstruction, for putting all cameras in same affine basis

  //::: find f-mats
  F.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) F[i][j]=FundamentalMatrix::NormalizedFind(Kpts[i],Kpts[j]);

  //::: find epipoles
  eR.SetSize(CON_ncameras,CON_ncameras);
  eL.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) FundamentalMatrix::FindEpis(eR[i][j],eL[i][j],F[i][j]);

  //::: find any M such that F == [eB]_x*M
  M.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) M[i][j]=Transform::Cross(eL[i][j])*F[i][j];  //ERROR!!!  tell Ziss!?
    if (i<j) FundamentalMatrix::FindNonhomogeneous(M[i][j],F[i][j],eL[i][j]);

  MakeInitialReconstructionRounder_2(one,two);


#else


//::: old path

  //::: find f-mats
  F.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) F[i][j]=FundamentalMatrix::NormalizedFind(Kpts[i],Kpts[j]);

  //::: find epipoles
  eR.SetSize(CON_ncameras,CON_ncameras);
  eL.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) FundamentalMatrix::FindEpis(eR[i][j],eL[i][j],F[i][j]);

  //::: find any M such that F == [eB]_x*M
  M.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) M[i][j]=Transform::Cross(eL[i][j])*F[i][j];  //ERROR!!!  tell Ziss!?
    if (i<j) FundamentalMatrix::FindNonhomogeneous(M[i][j],F[i][j],eL[i][j]);

//  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) { F[i][j].DumpNormalized(); (Transform::Cross(eL[i][j])*M[i][j]).DumpNormalized(); }

  //::: find true Hinf's
  Hinf.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) ExtractHinf(Hinf[i][j],CM[i],CM[j]);

  //::: find good initial reconstruction, for putting all cameras in same affine basis
  MakeInitialReconstructionRounder(one,two);

  //::: find alpha such that Hinf == M + e_B*alpha'
  VectorA.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) FindVectorA(VectorA[i][j],Hinf[i][j],M[i][j],eL[i][j]);
    // NOTE: must apply normalizing transform to camera matrices for FindVectorA to work

  //::: store vectorA in convenient form
  alpha.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j)
    alpha[i][j]=AlphaCoord(VectorA[i][j][0],VectorA[i][j][1],VectorA[i][j][2]);

  //::: pairwise projective reconstructions
  R.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if (i<j) MiscUtility::AffineTriangulate(R[i][j],M[i][j],F[i][j],Kpts[i],Kpts[j]);

  SCREENPUT VrmlFile::OutputPoints("output.initial.projective.recon.wrl",R[one][two],RGB3Byte_orange);

  //::: collineations between projective reconstructions
  T.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if ((i<j)&&((i!=0)||(j!=1))) MiscUtility::Calc4DCollineation(T[i][j],R[i][j],R[0][1]);
    if ((i<j)&&((i!=one)||(j!=two))) MiscUtility::Calc4DCollineation(T[i][j],R[one][two],R[i][j]);

  //::: find true parameterizations; store "view pairs" separately
  k.SetSize(CON_pairs);
  th.SetSize(CON_pairs);
  g.SetSize(CON_pairs);
  int ind=0;
  FORi(CON_ncameras) FORj(CON_ncameras) if (i<j) {
    FindParameterizationOfHMatrix(k[ind],th[ind],g[ind],internal,
      Hinf[i][j],F[i][j],M[i][j],eL[i][j],TRUE);
    ++ind;
  }
#endif
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
void FindPtsThatAreInAllViews(
  Array2D<rPoint3>&			Kpts,
  const KltFeatureTable&		ftab
)
{
  LinkedList<int> list;
  FORi(ftab.rows) {
    boole bOK=(ftab[i][0].msg>0);
    if (bOK) {
      for (int j=1; j<ftab[i].size; ++j)
        if (ftab[i][j].msg!=0) { bOK=FALSE; break; }
    }
    if (bOK) list.Add(new int(i));
  }
  Kpts.SetSize(ftab.cols,list.size);
  int inx=0;
  FORLISTnow(list,int)
    FORi(Kpts.rows) Kpts[i][inx]=rPoint3(ftab[now][i].x,ftab[now][i].y,1);
    ++inx;
  }
}

// constructor for real data (external data files)
///////////////////////////////////////////////////////////////////////
MultiCameraWorkbench::MultiCameraWorkbench(
  const String&			file_features,
  const String&			file_fmats,
  const int			one,
  const int			two
)
{
  //::: read features from file
  int imgRows, imgCols;
  KltFeatureTable ftab;
  ftab.ReadFromFile(imgRows,imgCols,file_features);
///ftab.ReadFromDenseMatchFile(imgRows,imgCols,"given.dense.matching");
///ALERT("check this....");

  CON_ncameras=ftab.cols;
  CON_noiseradius=0;
  CON_pairs=(CON_ncameras*(CON_ncameras-1))/2;

  FindPtsThatAreInAllViews(Kpts,ftab);
  CON_nworldpts=Kpts[0].size;
  printf("%d common pts\n",CON_nworldpts);
  if (Kpts[0].size==0) EXIT("not any common pts");
  if (Kpts[0].size<8) EXIT("not enough common pts");

  //::: normalize points on camera image planes
//  FORi(CON_ncameras) MiscUtility::FindNormalizingTransform(Tinorm[i],Kpts[i]);  //bad error! must normalize all cameras the same way
  MiscUtility::FindNormalizingTransform(normalization,Kpts[0]);

  FORi(Kpts.rows) FORj(Kpts.cols) {
    rPoint3& p=Kpts[i][j];
    p=normalization*p;				//note the z-coord remains the same (at 1)
  }

  //::: read f-mats from file; modify by normalizing transformations
//ORIGINAL!
  F.SetSize(CON_ncameras,CON_ncameras);
  bPairOn.SetSize(CON_ncameras,CON_ncameras);
  bPairOn.FillWith(FALSE);
  FILE* file=fopen(file_fmats.Ptr(),"r");
  if (!file) EXIT("couldn't open fmat file");
  Transform inv_normalization=normalization.Inverse();
  while (!feof(file)) {
    int i,j;
    fscanf(file,"%d %d",&i,&j);
    if (feof(file)) break;
    Transform& f=F[i][j];
    fscanf(file,"%lf %lf %lf",&f[0][0],&f[0][1],&f[0][2]);
    fscanf(file,"%lf %lf %lf",&f[1][0],&f[1][1],&f[1][2]);
    fscanf(file,"%lf %lf %lf",&f[2][0],&f[2][1],&f[2][2]);
    f=(inv_normalization).TransposeOf()*f*(inv_normalization);
    f=f.NormalizationOf();
    bPairOn[i][j]=TRUE;
      printf("%d  %d\n",i,j);
      f.Dump();
  }

/*
//::: KLUDGE!! find f-mats
F.SetSize(CON_ncameras,CON_ncameras);
FORi(CON_ncameras) FORj(CON_ncameras)
  if (i<j) F[i][j]=FundamentalMatrix::NormalizedFind(Kpts[i],Kpts[j]);
bPairOn.SetSize(CON_ncameras,CON_ncameras);
bPairOn.FillWith(TRUE);
FILE* file;
*/

/*
//::: sanity check on f-mats
FORi(CON_ncameras) FORj(CON_ncameras) if ((i<j)&&(bPairOn[i][j])) {
  BANNER("testing fmats");
  FORk(Kpts[i].size) {
    printf("%f\n",(Kpts[j][k].Hat()).Dot((F[i][j]*Kpts[i][k]).Hat()));
  }
}
*/



#ifdef NOTDEFINED
//::: new path here

  //::: find epipoles
  eR.SetSize(CON_ncameras,CON_ncameras);
  eL.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if ((i<j)&&(bPairOn[i][j])) FundamentalMatrix::FindEpis(eR[i][j],eL[i][j],F[i][j]);

  //::: find any M such that F == [eB]_x*M
  M.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) M[i][j]=Transform::Cross(eL[i][j])*F[i][j];  //ERROR!!!  tell Ziss!?
    if ((i<j)&&(bPairOn[i][j])) FundamentalMatrix::FindNonhomogeneous(M[i][j],F[i][j],eL[i][j]);

MakeInitialReconstructionRounder(one,two);
#endif



//::: old path
  //::: read full camera matrices (if they exist)
  file=fopen("given.camera.matrices","rb");
  if (file) {
    int nViews;
    fscanf(file,"%d\n",&nViews);
    CM.SetSize(nViews);
    rMatrix rm_normalization(&normalization[0][0],3,3);
    FORi(CM.size) {
      CM[i].SetSize(3,4);
      fscanf(file,"\n");
      FORj(3) fscanf(file,"%lf %lf %lf %lf\n",
        &CM[i][j][0],&CM[i][j][1],&CM[i][j][2],&CM[i][j][3]);
      CM[i]=rm_normalization*CM[i];	// apply normalizing transform
    }
    fclose(file);
    //::: find true internal calibration matrix from first camera matrix (assume would be the same for all other cam matrices)
    rMatrix KK(3,3);
    FORi(3) FORj(3) KK[i][j]=CM[0][i][j];
    KK=KK*(KK.TransposeOf());
    KK=KK.InverseOf();
    ASSERT(KK.Cholesky(),"Cholesky failed while trying to find true K");
    KK=KK.InverseOf();
    FORi(3) FORj(3) internal[i][j]=KK[i][j];
  }
  else internal=Transform(0,0,0,0,0,0,0,0,0);

  //::: find epipoles
  eR.SetSize(CON_ncameras,CON_ncameras);
  eL.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if ((i<j)&&(bPairOn[i][j])) FundamentalMatrix::FindEpis(eR[i][j],eL[i][j],F[i][j]);

  //::: find any M such that F == [eB]_x*M
  M.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) M[i][j]=Transform::Cross(eL[i][j])*F[i][j];  //ERROR!!!  tell Ziss!?
    if ((i<j)&&(bPairOn[i][j])) FundamentalMatrix::FindNonhomogeneous(M[i][j],F[i][j],eL[i][j]);

//  FORi(CON_ncameras) FORj(CON_ncameras)
//    if (i<j) { F[i][j].DumpNormalized(); (Transform::Cross(eL[i][j])*M[i][j]).DumpNormalized(); }

  // TO TRY:  FIND A FILLED-OUT PROJECTIVE RECONSTRUCTION, THEN
  // FIND ALL COMMON-BASIS CAMERA MATRICES FROM THIS, THEN(??) TRANSFORM
  // CAMERA BASES SO THAT FIRST ONE HAS FORM [ I | 0 ]
  //::: find good initial reconstruction, for putting all cameras in same affine basis
  MakeInitialReconstructionRounder(one,two);

  if (CM.size>0) {		// must have loaded the "true" camera matrices
    ASSERT(CM.size==CON_ncameras,"number of camera matrices in given file is wrong");
    //::: find true Hinf's
    Hinf.SetSize(CON_ncameras,CON_ncameras);
    FORi(CON_ncameras) FORj(CON_ncameras)
      if (i<j) ExtractHinf(Hinf[i][j],CM[i],CM[j]);

    //::: find alpha such that Hinf == M + e_B*alpha'
    VectorA.SetSize(CON_ncameras,CON_ncameras);
    FORi(CON_ncameras) FORj(CON_ncameras)
      if (i<j) FindVectorA(VectorA[i][j],Hinf[i][j],M[i][j],eL[i][j]);
      // NOTE: must apply normalizing transform to camera matrices for FindVectorA to work

    //::: store vectorA in convenient form
    alpha.SetSize(CON_ncameras,CON_ncameras);
    FORi(CON_ncameras) FORj(CON_ncameras) if (i<j)
      alpha[i][j]=AlphaCoord(VectorA[i][j][0],VectorA[i][j][1],VectorA[i][j][2]);
  }

  //::: pairwise projective reconstructions
  R.SetSize(CON_ncameras,CON_ncameras);
  FORi(CON_ncameras) FORj(CON_ncameras)
    if ((i<j)&&(bPairOn[i][j]))
      MiscUtility::AffineTriangulate(R[i][j],M[i][j].NormalizationOf(),F[i][j],Kpts[i],Kpts[j]);

  SCREENPUT VrmlFile::OutputPoints("initial.projective.recon.wrl",R[one][two],RGB3Byte_orange);

//  //::: transform projective reconstruction by a projective transformation so that
//  //::: principal components of point cloud are about equal

/*
  //::: turn off pairs without significant rotation between them
  FORi(CON_ncameras) FORj(CON_ncameras)
    if ((i<j)&&(bPairOn[i][j])) {
      const Transform& f=F[i][j];
      if (!((abs(f[0][0])>0.1)||(abs(f[1][1])>0.1)||(abs(f[2][2])>0.1))) //kludge to eliminate pairs that aren't rotated enough
        bPairOn[i][j]=FALSE;
    }
*/
}

